/**
 * Copyright (C) 2014 Roman Hiestand
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "CommonIncludes.h"
#include "PJDirTreeThread.h"
#include "PJFileType.h"

PJDirTreeThread::PJDirTreeThread() : wxThread(wxTHREAD_JOINABLE),
	jobType_(PJDTTJobTypeCompress), path_(wxEmptyString),
	mutex_()
{
}

PJDirTreeThread::~PJDirTreeThread()
{
}

bool PJDirTreeThread::startJob(PJDirTreeThread::PJDTTJobType jobType, const wxString &path)
{
	jobType_ = jobType;
	path_ = path;
	workFiles_.clear();

	if(Create() != wxTHREAD_NO_ERROR)
		return false;
	if(Run() != wxTHREAD_NO_ERROR)
		return false;

	return true;
}

bool PJDirTreeThread::getWorkFile(FoundFile &file)
{
	wxMutexLocker lock(mutex_);
	if(workFiles_.empty())
		return false;

	file = *(workFiles_.begin());
	workFiles_.pop_front();

	return true;
}

wxThread::ExitCode PJDirTreeThread::Entry()
{
	wxFileName inPath(path_, wxT("")), outPath;
	outPath = inPath;

	// Append _compressed to last directory name
	wxArrayString pathnames = outPath.GetDirs();
	if(pathnames.size() > 0)
	{
		outPath.RemoveLastDir();
		if(jobType_ == PJDirTreeThread::PJDTTJobTypeCompress)
			outPath.AppendDir(pathnames[pathnames.size() - 1] + _T("_compressed"));
		else
			outPath.AppendDir(pathnames[pathnames.size() - 1] + _T("_uncompressed"));
	}
	traverseDir(inPath, outPath);

	return 0;
}

void PJDirTreeThread::traverseDir(const wxFileName &path, const wxFileName &outpath)
{
	wxString currentPathName(path.GetPath());
	wxDir curDir(currentPathName);
	if(curDir.IsOpened())
	{
		wxString filename;
#if wxCHECK_VERSION(2, 9, 0)
		bool hasMore = curDir.GetFirst(&filename, wxEmptyString, wxDIR_FILES | wxDIR_DIRS | wxDIR_NO_FOLLOW);
#else
		bool hasMore = curDir.GetFirst(&filename, wxEmptyString, wxDIR_FILES | wxDIR_DIRS);
#endif
		while(hasMore)
		{
			wxFileName curItemFile(currentPathName, filename);
			wxFileName curItemDir(path);
			curItemDir.AppendDir(filename);

			if(curItemFile.FileExists())	// It's a file
			{
				wxString fullFilename = curItemFile.GetFullPath();
				wxFileName outFilename(outpath);
				outFilename.SetFullName(filename);
				if(jobType_ == PJDirTreeThread::PJDTTJobTypeCompress)
					outFilename.SetExt(wxT("pjg"));
				else
					outFilename.SetExt(wxT("jpg"));

				FoundFile file;
				file.inFilename = fullFilename;
				file.outFilename = outFilename.GetFullPath();
				PJFileType::FileType fileType = PJFileType::determineFileType(fullFilename);
				if(jobType_ == PJDirTreeThread::PJDTTJobTypeCompress && fileType == PJFileType::FileTypeJPEG)
				{
					addWorkFile(file);
				}
				else if(jobType_ == PJDirTreeThread::PJDTTJobTypeDecompress && fileType == PJFileType::FileTypePJG)
				{
					addWorkFile(file);
				}
			}
			else if(curItemDir.DirExists())	// It's a directory
			{
				// Call myself recursively
				wxFileName newOutPath(outpath);
				newOutPath.AppendDir(filename);
				traverseDir(curItemDir, newOutPath);
			}
			hasMore = curDir.GetNext(&filename);
		}
	}
}

void PJDirTreeThread::addWorkFile(const FoundFile &file)
{
	wxMutexLocker lock(mutex_);
	workFiles_.push_back(file);
}
