/**
 * Copyright (C) 2014 Roman Hiestand
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "CommonIncludes.h"
#include "PJDropTargetWin32.h"
#include "wxPackJPGMainFrame.h"
#include "wxPackJPGMainFrameBase.h"


// The following code is inspired by (partly copied from) Raymond Chen, see
// his blog on "How do I accept files to be opened via IDropTarget instead of on the command line?", link:
// http://blogs.msdn.com/b/oldnewthing/archive/2010/05/03/10006065.aspx

#include <shlobj.h>
#include <shellapi.h>
#include <docobj.h>
#include <new> // for new(nothrow)

class ProcessReference : public IUnknown {
public:
	STDMETHODIMP QueryInterface(REFIID riid, void **ppv)
	{
		if (riid == IID_IUnknown) {
			*ppv = static_cast<IUnknown*>(this);
			AddRef();
			return S_OK;
		}
		*ppv = NULL; return E_NOINTERFACE;
	}

	STDMETHODIMP_(ULONG) AddRef()
	{ return InterlockedIncrement(&m_cRef); }

	STDMETHODIMP_(ULONG) Release()
	{
		LONG lRef = InterlockedDecrement(&m_cRef);
		if (lRef == 0) PostThreadMessage(m_dwThread, WM_NULL, 0, 0);
		return lRef;
	}

	ProcessReference()
		: m_cRef(1), m_dwThread(GetCurrentThreadId())
	{ }

	~ProcessReference()
	{
		Release();

		MSG msg;
		while (m_cRef && GetMessage(&msg, NULL, 0, 0))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

private:
	LONG m_cRef;
	DWORD m_dwThread;
};

ProcessReference g_pr;
ProcessReference *g_ppr = &g_pr;

class PJOLEDropTarget : public IDropTarget, public IOleCommandTarget
{
public:
	PJOLEDropTarget() : m_cRef(1), pPJDropTargetWin32_(NULL) { g_ppr->AddRef(); }
	~PJOLEDropTarget() { g_ppr->Release(); }

	// *** IUnknown ***
	STDMETHODIMP QueryInterface(REFIID riid, void **ppv)
	{
		*ppv = NULL;
		if (riid == IID_IUnknown)
			*ppv = (IDropTarget*)this;
		else if(riid == IID_IDropTarget)
			*ppv = (IDropTarget*)this;
		else if(riid == IID_IOleCommandTarget)
			*ppv = (IOleCommandTarget*)this;
		else
			return E_NOINTERFACE;
		AddRef();
		return S_OK;
	}

	STDMETHODIMP_(ULONG) AddRef()
	{
		return InterlockedIncrement(&m_cRef);
	}

	STDMETHODIMP_(ULONG) Release()
	{
		LONG cRef = InterlockedDecrement(&m_cRef);
		if (cRef == 0) delete this;
		return cRef;
	}

	// *** IDropTarget ***
	STDMETHODIMP DragEnter(IDataObject *pdto,
		DWORD grfKeyState, POINTL ptl, DWORD *pdwEffect)
	{
		*pdwEffect &= DROPEFFECT_COPY;
		return S_OK;
	}

	STDMETHODIMP DragOver(DWORD grfKeyState,
		POINTL ptl, DWORD *pdwEffect)
	{
		*pdwEffect &= DROPEFFECT_COPY;
		return S_OK;
	}

	STDMETHODIMP DragLeave()
	{
		return S_OK;
	}

	STDMETHODIMP Drop(IDataObject *pdto, DWORD grfKeyState,
		POINTL ptl, DWORD *pdwEffect)
	{
		pPJDropTargetWin32_->OpenFilesFromDataObject(pdto);
		*pdwEffect &= DROPEFFECT_COPY;
		return S_OK;
	}

	// *** IOleCommandTarget ***
	virtual STDMETHODIMP QueryStatus( 
		const GUID *pguidCmdGroup,
		ULONG cCmds,
		OLECMD prgCmds[  ],
		OLECMDTEXT *pCmdText)
	{
		return S_OK;
	}

	virtual STDMETHODIMP Exec( 
		const GUID *pguidCmdGroup,
		DWORD nCmdID,
		DWORD nCmdexecopt,
		VARIANT *pvaIn,
		VARIANT *pvaOut)
	{
		if(nCmdID == wxPJCmd_Transfer_Filename && pvaIn != NULL)
		{
			if(pvaIn->vt == VT_BSTR)
			{
				wxString inStr(pvaIn->bstrVal);
				pPJDropTargetWin32_->addFile(inStr);
			}
			else if(pvaIn->vt == (VT_ARRAY | VT_BSTR))
			{
				VARTYPE vt;
				HRESULT hr = SafeArrayGetVartype(pvaIn->parray, &vt);
				if(SafeArrayGetDim(pvaIn->parray) == 1
					&& SUCCEEDED(hr) && vt == VT_BSTR)
				{
					// Get array bounds
					LONG lowerBound, upperBound;
					SafeArrayGetLBound(pvaIn->parray, 1, &lowerBound);
					SafeArrayGetUBound(pvaIn->parray, 1, &upperBound);

					// Directly access safe array
					BSTR *bstrVal;
					wxArrayString filenames;
					hr = SafeArrayAccessData(pvaIn->parray, (void**)&bstrVal);
					long count = upperBound - lowerBound + 1; 
					for (int i = 0; i < count; i++)
					{
						BSTR curBstr = bstrVal[i];
						wxString curStr(curBstr);
						filenames.Add(curBstr);
					}
					SafeArrayUnaccessData(pvaIn->parray);

					pPJDropTargetWin32_->addFiles(filenames);
				}
			}
		}
		else if(nCmdID == wxPJCmd_Compress_Here)
		{
			pPJDropTargetWin32_->startCommand(wxPJCmd_Compress_Here);
		}
		else if(nCmdID == wxPJCmd_Decompress_Here)
		{
			pPJDropTargetWin32_->startCommand(wxPJCmd_Decompress_Here);
		}
		else if(nCmdID == wxPJCmd_Compress_To_Folder && pvaIn != NULL && pvaIn->vt == VT_BSTR)
		{
			wxString pathStr(pvaIn->bstrVal);
			pPJDropTargetWin32_->startCommand(wxPJCmd_Compress_To_Folder, pathStr);
		}
		else if(nCmdID == wxPJCmd_Decompress_To_Folder && pvaIn != NULL && pvaIn->vt == VT_BSTR)
		{
			wxString pathStr(pvaIn->bstrVal);
			pPJDropTargetWin32_->startCommand(wxPJCmd_Decompress_To_Folder, pathStr);
		}
		else if(nCmdID == wxPJCmd_Compress_Folder_To_PJG && pvaIn != NULL && pvaIn->vt == VT_BSTR)
		{
			wxString pathStr(pvaIn->bstrVal);
			pPJDropTargetWin32_->startCommand(wxPJCmd_Compress_Folder_To_PJG, pathStr);
		}
		else if(nCmdID == wxPJCmd_Decompress_Folder_To_JPG && pvaIn != NULL && pvaIn->vt == VT_BSTR)
		{
			wxString pathStr(pvaIn->bstrVal);
			pPJDropTargetWin32_->startCommand(wxPJCmd_Decompress_Folder_To_JPG, pathStr);
		}
		else if(nCmdID == wxPJCmd_Get_HWND && pvaOut != NULL)
		{
			WXHWND hwnd = wxTheApp->GetTopWindow()->GetHWND();
			VariantChangeType(pvaOut, pvaOut, VARIANT_NOVALUEPROP, VT_UI8);
			pvaOut->ullVal = (ULONGLONG)hwnd;
		}

		return S_OK;
	}

	void setPJDropTargetWin32(PJDropTargetWin32 *pPJDropTargetWin32)
	{
		pPJDropTargetWin32_ = pPJDropTargetWin32;
	}

private:
	LONG m_cRef;
	PJDropTargetWin32 *pPJDropTargetWin32_;
};

class PJOLEClassFactory : public IClassFactory
{
public:
	// *** IUnknown ***
	STDMETHODIMP QueryInterface(REFIID riid, void **ppv)
	{
		if (riid == IID_IUnknown || riid == IID_IClassFactory)
		{
			*ppv = static_cast<IUnknown*>(this);
			AddRef();
			return S_OK;
		}
		*ppv = NULL;
		return E_NOINTERFACE;
	}

	STDMETHODIMP_(ULONG) AddRef()
	{
		return 2;
	}

	STDMETHODIMP_(ULONG) Release()
	{
		return 1;
	}

	// *** IClassFactory ***
	STDMETHODIMP CreateInstance(IUnknown *pUnkOuter, REFIID riid, void **ppv)
	{
		*ppv = NULL;
		if (pUnkOuter) return CLASS_E_NOAGGREGATION;
		HRESULT hr = E_NOINTERFACE;
		if(riid == IID_IUnknown || riid == IID_IDropTarget || riid == IID_IOleCommandTarget)
		{
			PJOLEDropTarget *pdt = new(std::nothrow) PJOLEDropTarget();
			if (!pdt) return E_OUTOFMEMORY;
			hr = pdt->QueryInterface(riid, ppv);
			pdt->setPJDropTargetWin32(pPJDropTargetWin32_);
			pdt->Release();
		}

		return hr;
	}

	STDMETHODIMP LockServer(BOOL fLock)
	{
		if (!g_ppr) return E_FAIL; // server shutting down
		if (fLock) g_ppr->AddRef(); else g_ppr->Release();
		return S_OK;
	}

	void setPJDropTargetWin32(PJDropTargetWin32 *pPJDropTargetWin32)
	{
		pPJDropTargetWin32_ = pPJDropTargetWin32;
	}

private:
	PJDropTargetWin32 *pPJDropTargetWin32_;
};

PJDropTargetWin32::PJDropTargetWin32():
	hrRegister_(0), dwRegisterCookie_(0)
{
	pPJOCF_ = new PJOLEClassFactory();
	pPJOCF_->setPJDropTargetWin32(this);
	hrRegister_ = CoRegisterClassObject(CLSID_WXPACKJPG, pPJOCF_,
		CLSCTX_LOCAL_SERVER, REGCLS_MULTIPLEUSE,
		&dwRegisterCookie_);
}

PJDropTargetWin32::~PJDropTargetWin32()
{
	if (SUCCEEDED(hrRegister_))
	{
		CoRevokeClassObject(dwRegisterCookie_);
	}
	delete pPJOCF_;
}

void PJDropTargetWin32::setMainFrame(wxPackJPGMainFrame *pWxPackJPGMainFrame)
{
	pMainFrame_ = pWxPackJPGMainFrame;
}

void PJDropTargetWin32::OpenFilesFromDataObject(IDataObject *pdto)
{
	wxArrayString filenames;
	FORMATETC fmte = { CF_HDROP, NULL, DVASPECT_CONTENT,
		-1, TYMED_HGLOBAL };
	STGMEDIUM stgm;
	if (SUCCEEDED(pdto->GetData(&fmte, &stgm)))
	{
		HDROP hdrop = reinterpret_cast<HDROP>(stgm.hGlobal);
		UINT cFiles = DragQueryFile(hdrop, 0xFFFFFFFF, NULL, 0);
		for (UINT i = 0; i < cFiles; i++)
		{
			TCHAR szFile[MAX_PATH];
			UINT cch = DragQueryFile(hdrop, i, szFile, MAX_PATH);
			if (cch > 0 && cch < MAX_PATH)
			{
				filenames.Add(wxString(szFile));
			}
		}
		ReleaseStgMedium(&stgm);
	}

	if(pMainFrame_ != NULL)
	{
		pMainFrame_->addShellExtFiles(filenames);
	}
}

void PJDropTargetWin32::addFile(const wxString &filename)
{
	wxArrayString filenames;

	filenames.Add(filename);
	if(pMainFrame_ != NULL)
	{
		pMainFrame_->addShellExtFiles(filenames);
	}
}

void PJDropTargetWin32::addFiles(const wxArrayString &filenames)
{
	if(pMainFrame_ != NULL)
	{
		pMainFrame_->addShellExtFiles(filenames);
	}
}

void PJDropTargetWin32::startCommand(wxPackJPG_OLE_Commands cmd, const wxString &pathStr)
{
	if(pMainFrame_ != NULL)
	{
		pMainFrame_->startCommand(cmd, pathStr);
	}
}

void PJDropTargetWin32::registerServer()
{
	LPWSTR clsidString = NULL;
	wchar_t subKey[1024];
	if(FAILED(StringFromIID(CLSID_WXPACKJPG, &clsidString)))
		return;
	wsprintf(subKey, _T("CLSID\\%s\\LocalServer32"), clsidString);
	CoTaskMemFree(clsidString);

	HKEY result;
	LONG retVal = RegCreateKeyEx(HKEY_CLASSES_ROOT, subKey, 0, NULL,
		REG_OPTION_NON_VOLATILE, KEY_WRITE | KEY_WOW64_64KEY, NULL, &result, NULL);
	if(retVal == ERROR_SUCCESS)
	{
		WCHAR moduleFilename[MAX_PATH];
		GetModuleFileName(NULL, moduleFilename, MAX_PATH);

		retVal = RegSetValueEx(result, NULL, 0, REG_SZ, reinterpret_cast<const BYTE *>(moduleFilename), 
			sizeof(moduleFilename[0])*(wcslen(moduleFilename) + 1));
		RegCloseKey(result);
	}

}

void PJDropTargetWin32::unregisterServer()
{
	LPWSTR clsidString = NULL;
	wchar_t subKeyLS[1024], subKey[1024];
	if(FAILED(StringFromIID(CLSID_WXPACKJPG, &clsidString)))
		return;
	wsprintf(subKeyLS, _T("CLSID\\%s\\LocalServer32"), clsidString);
	wsprintf(subKey, _T("CLSID\\%s"), clsidString);
	CoTaskMemFree(clsidString);

	// Since Windows XP does not know this function, use GetProcAddress
	typedef LONG (APIENTRY *LPFN_REGDELETEKEYEXW) (HKEY, LPCWSTR, REGSAM, DWORD);
	LPFN_REGDELETEKEYEXW fnRegDeleteKeyExW;
	fnRegDeleteKeyExW = (LPFN_REGDELETEKEYEXW) GetProcAddress(
		GetModuleHandle(TEXT("advapi32")),"RegDeleteKeyExW");
	if(fnRegDeleteKeyExW != NULL)
	{
		LONG retVal = fnRegDeleteKeyExW(HKEY_CLASSES_ROOT, subKeyLS, KEY_WOW64_64KEY, 0);
		retVal = fnRegDeleteKeyExW(HKEY_CLASSES_ROOT, subKey, KEY_WOW64_64KEY, 0);
	}
	else
	{
		LONG retVal = RegDeleteKey(HKEY_CLASSES_ROOT, subKeyLS);
		retVal = RegDeleteKey(HKEY_CLASSES_ROOT, subKey);
	}
}
