/**
 * Copyright (C) 2014 Roman Hiestand
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "CommonIncludes.h"
#include "PJWorkerThread.h"
#include "PJWorkerThreadManager.h"
#include "packJPGlib/packjpglib.h"

wxMutex PJWorkerThread::fileSystemLocker_;

PJWorkerThread::PJWorkerThread() : wxThread(wxTHREAD_JOINABLE),
	pPJWorkerThreadManager_(NULL), verify_(false),
	ignoreWarnings_(false), discardMetadata_(false),
	createDir_(false)
{
}

PJWorkerThread::~PJWorkerThread()
{
}

void PJWorkerThread::setWorkerThreadManager(PJWorkerThreadManager *pPJWorkerThreadManager)
{
	pPJWorkerThreadManager_ = pPJWorkerThreadManager;
}

void PJWorkerThread::setParameters(bool verify, bool ignoreWarnings, bool discardMetadata, bool createDir)
{
	verify_ = verify;
	ignoreWarnings_ = ignoreWarnings;
	discardMetadata_ = discardMetadata;
	createDir_ = createDir;
}

wxThread::ExitCode PJWorkerThread::Entry()
{
	wxString inFilename, outFilename;
	bool hasWorkItem = pPJWorkerThreadManager_->getWorkItem(inFilename, outFilename);
	while(hasWorkItem)
	{
		char msg[1024];
		bool isOK = true;

		if(createDir_)
		{
			wxFileName fn(outFilename);
			fn.SetFullName(wxEmptyString);	// Use only path part

			// The following block has a mutex to prevent another thread trying to
			// create the same directory at the same time.
			{
				wxMutexLocker lock(fileSystemLocker_);
				if(!fn.DirExists())
#if wxCHECK_VERSION(2, 9, 0)
					fn.Mkdir(wxS_DIR_DEFAULT, wxPATH_MKDIR_FULL);	// Also create parent directories if necessary
#else
					fn.Mkdir(0777, wxPATH_MKDIR_FULL);	// Also create parent directories if necessary
#endif
			}
		}

		if(verify_)
		{
			// Read file to memory buffer
			wxFFile infile(inFilename, wxT("rb"));
			if(!infile.IsOpened())
			{
				strcpy(msg, "Input file could not be opened");
				isOK = false;
			}
			else
			{
				wxFileOffset fileSize = infile.Length();
				void *filebuf = malloc(fileSize);
				infile.Read(filebuf, fileSize);
				infile.Close();

				// Pack JPEG
				void *wa = pjglib_alloc_work_area();
				pjglib_adjust_settings((ignoreWarnings_ ? 2 : 1), discardMetadata_, wa);
				pjglib_init_streams( filebuf, 1, static_cast<int>(fileSize), NULL, 1, wa );

				unsigned char *out_file = NULL, *out_file_up = NULL;
				unsigned int out_size = 0, out_size_up = 0;
				bool ret_val = pjglib_convert_stream2mem( &out_file, &out_size, msg, wa );

				pjglib_free_work_area( wa );

				if(ret_val)
				{
					// Unpack again
					wa = pjglib_alloc_work_area();
					pjglib_adjust_settings((ignoreWarnings_ ? 2 : 1), discardMetadata_, wa);
					pjglib_init_streams( out_file, 1, out_size, NULL, 1, wa );

					char msg_up[1024];
					bool ret_val_up = pjglib_convert_stream2mem( &out_file_up, &out_size_up, msg_up, wa );

					pjglib_free_work_area( wa );

					if(ret_val_up)
					{
						// Compare original file and unpacked file
						bool verifySuccessful = true;
						if(out_size_up == static_cast<unsigned int>(fileSize))
						{
							unsigned char* origFile = reinterpret_cast<unsigned char*>(filebuf);
							for(unsigned int i = 0; i < out_size_up; i++)
							{
								verifySuccessful &= (origFile[i] == out_file_up[i]);
							}
						}

						// Write packed JPEG to file
						if(verifySuccessful)
						{
							wxFFile outfile(outFilename, wxT("wb"));
							if(!outfile.IsOpened())
							{
								strcpy(msg, "Output file could not be opened for writing");
								isOK = false;
							}
							else
							{
								outfile.Write(out_file, out_size);
								outfile.Close();
							}
						}
						else
						{
							strcpy(msg, "Verify failed");
							isOK = false;
						}
					}
					else
					{
						strcpy(msg, msg_up);
						isOK = false;
					}
				}
				else
				{
					isOK = false;
				}

				// Free pointers returned from packJPGlib
				free(out_file_up);
				free(out_file);

				free(filebuf);
			}
		}
		else
		{
			void *wa = pjglib_alloc_work_area();
			pjglib_adjust_settings((ignoreWarnings_ ? 2 : 1), discardMetadata_, wa);
			isOK = pjglib_convert_file2file( const_cast<char *>((const char *)(inFilename.mb_str( *wxConvCurrent ))),
				const_cast<char *>((const char *)(outFilename.mb_str( *wxConvCurrent ))), msg, wa );
			pjglib_free_work_area( wa );
		}

		if(!isOK)
		{
			// Remove outfile if it was created
			wxFileName ofn(outFilename);
			if(ofn.FileExists())
				wxRemoveFile(ofn.GetFullPath());
		}

		wxString processMsg(msg, *wxConvCurrent);
		pPJWorkerThreadManager_->finishedWorkItem(inFilename, outFilename, processMsg, isOK);

		hasWorkItem = pPJWorkerThreadManager_->getWorkItem(inFilename, outFilename);
	}

	return 0;
}
