/**
 * Copyright (C) 2014 Roman Hiestand
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "CommonIncludes.h"
#include "PJWorkerThreadManager.h"
#include "PJWorkerThread.h"

PJWorkerThreadManager::PJWorkerThreadManager() :
	mutex_(), isWorking_(false), verify_(false),
	ignoreWarnings_(false), discardMetadata_(false), createDir_(false),
	todoWorkItemsCount_(0), readFinishedWorkItemIndex_(0)
{
	todoWorkItems_.clear();
	finishedWorkItems_.clear();
	workerThreads_.clear();
}

PJWorkerThreadManager::~PJWorkerThreadManager()
{
}

void PJWorkerThreadManager::clearWorkItems()
{
	wxMutexLocker lock(mutex_);
	todoWorkItems_.clear();
	finishedWorkItems_.clear();
	readFinishedWorkItemIndex_ = 0;
}

void PJWorkerThreadManager::addWorkItem(wxString inFilename, wxString outFilename)
{
	wxMutexLocker lock(mutex_);
	WorkItem item;
	item.inFilename = inFilename;
	item.outFilename = outFilename;
	todoWorkItems_.push_back(item);
}

bool PJWorkerThreadManager::existsOutfilename(const wxString &outFilename)
{
	wxFileName ofn1(outFilename);

	wxMutexLocker lock(mutex_);
	std::list<WorkItem>::const_iterator iter = todoWorkItems_.begin();
	while(iter != todoWorkItems_.end())
	{
#if defined(PJ_WIN32)
		if(outFilename.IsSameAs(iter->outFilename, false))
			return true;
#else
		wxFileName ofn2(iter->outFilename);
		if(ofn1.SameAs(ofn2))
			return true;
#endif
		iter++;
	}

	return false;
}

void PJWorkerThreadManager::setParameters(bool verify, bool ignoreWarnings, bool discardMetadata, bool createDir)
{
	verify_ = verify;
	ignoreWarnings_ = ignoreWarnings;
	discardMetadata_ = discardMetadata;
	createDir_ = createDir;
}

void PJWorkerThreadManager::startWorking()
{
	wxMutexLocker lock(mutex_);
	todoWorkItemsCount_ = todoWorkItems_.size();
	finishedWorkItems_.clear();
	finishedWorkItems_.reserve(todoWorkItemsCount_);
	readFinishedWorkItemIndex_ = 0;
	isWorking_ = true;

	const int numberOfWorkerThreads = wxThread::GetCPUCount();
	for(int i = 0; i < numberOfWorkerThreads; i++)
	{
		PJWorkerThread *pPJWorkerThread = new PJWorkerThread();
		pPJWorkerThread->setWorkerThreadManager(this);
		pPJWorkerThread->setParameters(verify_, ignoreWarnings_, discardMetadata_, createDir_);
		pPJWorkerThread->Create();
		pPJWorkerThread->Run();

		workerThreads_.push_back(pPJWorkerThread);
	}
}

double PJWorkerThreadManager::getProgress()
{
	wxMutexLocker lock(mutex_);
	double totalItems = static_cast<double>(todoWorkItemsCount_);
	double finishedItems = static_cast<double>(finishedWorkItems_.size());
	double itemsInProgress = static_cast<double>(todoWorkItemsCount_ - finishedWorkItems_.size() - todoWorkItems_.size());

	return ( (finishedItems + 0.5 * itemsInProgress) / totalItems );
}

bool PJWorkerThreadManager::isFinishedWorking()
{
	wxMutexLocker lock(mutex_);
	bool isFinished = (todoWorkItemsCount_ == finishedWorkItems_.size());

	return isFinished;
}

void PJWorkerThreadManager::stopWorking()
{
	size_t numberOfWorkerThreads = workerThreads_.size();
	for(size_t i = 0; i < numberOfWorkerThreads; i++)
	{
		PJWorkerThread *pPJWorkerThread = workerThreads_[i];
		pPJWorkerThread->Wait();
		delete pPJWorkerThread;
	}

	workerThreads_.clear();
	isWorking_ = false;
}

bool PJWorkerThreadManager::readFinishedWorkItem(wxString &inFilename, wxString &outFilename,
	wxString &processMsg, bool &isOK)
{
	wxMutexLocker lock(mutex_);

	if(readFinishedWorkItemIndex_ >= finishedWorkItems_.size())
		return false;

	WorkItem &item = finishedWorkItems_[readFinishedWorkItemIndex_];
	inFilename = item.inFilename;
	outFilename = item.outFilename;
	processMsg = item.processMsg;
	isOK = item.isOK;
	readFinishedWorkItemIndex_++;

	return true;
}

/**
 * Remove all remaining work items from the todo list.
 */
void PJWorkerThreadManager::cancelWorking()
{
	wxMutexLocker lock(mutex_);

	todoWorkItemsCount_ -= todoWorkItems_.size();
	todoWorkItems_.clear();
}

bool PJWorkerThreadManager::getWorkItem(wxString &inFilename, wxString &outFilename)
{
	wxMutexLocker lock(mutex_);

	if(todoWorkItems_.size() == 0)
		return false;

	WorkItem item = todoWorkItems_.front();
	todoWorkItems_.pop_front();
	inFilename = item.inFilename;
	outFilename = item.outFilename;

	return true;
}

void PJWorkerThreadManager::finishedWorkItem(wxString inFilename, wxString outFilename, wxString processMsg, bool isOK)
{
	wxMutexLocker lock(mutex_);

	WorkItem item;
	item.inFilename = inFilename;
	item.outFilename = outFilename;
	item.processMsg = processMsg;
	item.isOK = isOK;
	finishedWorkItems_.push_back(item);
}
