/**
 * Copyright (C) 2014 Roman Hiestand
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "CommonIncludes.h"
#include "PreviewGeneratorThread.h"
#include "PJFileType.h"
#include "packJPGlib/packjpglib.h"
#include <wx/mstream.h>

PreviewGeneratorThread::PreviewGeneratorThread()
	: wxThread(wxTHREAD_JOINABLE),
	mutex_(), cond_(mutex_), previewRequestFilename_(wxEmptyString),
	hasNewPreviewRequest_(false), terminate_(false), cancelPreview_(false),
	outputMutex_(), previewImage_(), previewFilename_(wxEmptyString),
	hasNewPreview_(false)
{
}

PreviewGeneratorThread::~PreviewGeneratorThread()
{
}

bool PreviewGeneratorThread::startPreviewThread()
{
	wxThreadError err = Create();
	if(err != wxTHREAD_NO_ERROR)
		return false;

	err = Run();
	if(err != wxTHREAD_NO_ERROR)
		return false;

	return true;
}

void PreviewGeneratorThread::stopPreviewThread()
{
	{
		wxMutexLocker lock(mutex_);
		terminate_ = true;
		cond_.Broadcast();
	}
	wxThread::ExitCode code = Wait();
}

// Adding a new preview request
void PreviewGeneratorThread::addPreviewRequest(const wxString &filename)
{
	{
		wxMutexLocker lock(outputMutex_);

		cancelPreview_ = false;
	}

	{
		wxMutexLocker lock(mutex_);

		previewRequestFilename_ = filename;
		hasNewPreviewRequest_ = true;
		cond_.Signal();
	}
}

// Ask for finished preview image
bool PreviewGeneratorThread::getNewPreviewImage(wxImage &previewImage, wxString &previewFilename)
{
	wxMutexLocker lock(outputMutex_);

	if(!hasNewPreview_)
		return false;

	if(cancelPreview_)
	{
		cancelPreview_ = false;
		hasNewPreview_ = false;
		return false;
	}

	previewImage = previewImage_;
	previewFilename = previewFilename_;
	hasNewPreview_ = false;

	return true;
}

void PreviewGeneratorThread::cancelPreview()
{
	wxMutexLocker lock(outputMutex_);

	cancelPreview_ = true;
}

wxThread::ExitCode PreviewGeneratorThread::Entry()
{
	while(!TestDestroy() && !terminate_)
	{
		wxImage localPreviewImage;
		wxString localPreviewRequestFilename;

		// Wait for new job
		{
			wxMutexLocker lock(mutex_);

			while(!hasNewPreviewRequest_
				&&!terminate_)
			{
				cond_.Wait();
			}

			localPreviewRequestFilename = previewRequestFilename_;
			hasNewPreviewRequest_ = false;
		}

		if(terminate_)
			return 0;

		bool fileReadOK = true;

		try
		{
			fileReadOK = createPreview(localPreviewRequestFilename, localPreviewImage);
		}
		catch(std::exception &e)
		{
			fileReadOK = false;
		}
		catch(...)
		{
			fileReadOK = false;
		}

		// Copy preview to previewImage_
		if(fileReadOK)
		{
			wxMutexLocker lock(outputMutex_);
			previewImage_ = localPreviewImage.Copy();	// Make an actual copy, not just a new reference
			previewFilename_ = localPreviewRequestFilename;
			hasNewPreview_ = true;
		}

	}

	return 0;
}


bool PreviewGeneratorThread::createPreview(const wxString &localPreviewRequestFilename, wxImage &localPreviewImage)
{
	bool fileReadOK = false;

	PJFileType::FileType fileType = PJFileType::determineFileType(localPreviewRequestFilename);
	if(fileType == PJFileType::FileTypePJG)
	{
		// Decompress from Packed JPG

		// Read file to memory buffer
		wxFFile infile(localPreviewRequestFilename, wxT("rb"));
		if(!infile.IsOpened())
			return false;

		wxFileOffset fileSize = infile.Length();
		void *filebuf = malloc(fileSize);
		if(filebuf == NULL)
			return false;
		infile.Read(filebuf, fileSize);
		infile.Close();

		void *wa = pjglib_alloc_work_area();
		pjglib_adjust_settings(1, false, wa);
		pjglib_init_streams( filebuf, 1, static_cast<int>(fileSize), NULL, 1, wa );
		unsigned char *out_file = NULL;
		unsigned int out_size = 0;
		char msg_up[1024];
		bool ret_val = pjglib_convert_stream2mem( &out_file, &out_size, msg_up, wa );

		pjglib_free_work_area( wa );

		if(!ret_val
			|| out_file == NULL)
		{
			free(filebuf);
			return false;
		}

		// Load JPEG image from memory
		wxMemoryInputStream memStream(out_file, out_size);
		fileReadOK = localPreviewImage.LoadFile(memStream, wxBITMAP_TYPE_JPEG);

		free(out_file);
		free(filebuf);
	}
	else if(fileType == PJFileType::FileTypeJPEG)
	{
		// Temporarily switch off logging errors for component wx
		// This is to prevent telling the user in a pop-up window that loading a
		// malformed JPEG for the preview did not work
		wxLogLevel prevLevel = wxLog::GetLogLevel();
#if wxCHECK_VERSION(2, 9, 0)
		wxLog::SetComponentLevel(wxT("wx"), wxLOG_FatalError);
#else
		wxLog::SetLogLevel(wxLOG_FatalError);
#endif
		// Load JPEG image
		fileReadOK = localPreviewImage.LoadFile(localPreviewRequestFilename, wxBITMAP_TYPE_JPEG);

#if wxCHECK_VERSION(2, 9, 0)
		wxLog::SetComponentLevel(wxT("wx"), prevLevel);
#else
		wxLog::SetLogLevel(prevLevel);
#endif
	}

	return fileReadOK;
}
