/**
 * Copyright (C) 2014 Roman Hiestand
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "CommonIncludes.h"

#include "wxPackJPGMainFrame.h"
#include "version.h"
#include "pj_config.h"
#include "PJDropTarget.h"
#if defined(PJ_WIN32)
#include "PJDropTargetWin32.h"
#endif
#include "packJPGlib/packjpglib.h"
#include "PJWorkerThreadManager.h"
#include "PJOverwriteDialog.h"
#include "PJFileType.h"
#include "PJDirTreeThread.h"

#if defined(__WXMAC__) || defined(__WXOSX__) || defined(__WXOSX_COCOA__)
#	include "osx/MacMenuWorkaroundBridge.h"
#endif

enum
{
	ID_TIMER = 2000,
	ID_SHOW_FILETYPE_WARNING_JPEG,
	ID_SHOW_FILETYPE_WARNING_PJG,
	ID_SHOW_PROCESSING_ERRORS,
	ID_CMD_COMPRESS_FOLDER_TO_PJG,
	ID_CMD_DECOMPRESS_FOLDER_TO_JPG,
	ID_START_TO_FOLDER
};

const wxEventType myCustomEventType = wxNewEventType();

// Static data members
int wxPackJPGMainFrame::sortColumn_ = 0;
wxListCtrl *wxPackJPGMainFrame::pSortListCtrl_ = NULL;
std::vector<size_t> wxPackJPGMainFrame::sortFileSizesC_, wxPackJPGMainFrame::sortFileSizesUC_, *wxPackJPGMainFrame::pSortFileSizes_ = NULL;
int wxPackJPGMainFrame::sortDirections_[4];
int wxPackJPGMainFrame::sortDirectionOffset_ = 0;


wxPackJPGMainFrame::wxPackJPGMainFrame(wxWindow* parent, bool calledFromCOM_OLE)
	: wxPackJPGMainFrameBase(parent),
	aTimer_(this, ID_TIMER),
	pDropTarget_(NULL), pPJDropTargetWin32_(NULL), pPJWorkerThreadManager_(NULL),
	failedItemsCount_(0), warningItemsCount_(0), pPJDirTreeThread_(NULL),
	firstTimeOnTimer_(true), calledFromCOM_OLE_(calledFromCOM_OLE),
	closeAfterCommandFinished_(false)
{
	nextListID_[0] = nextListID_[1] = 0;
	sortDirections_[0] = sortDirections_[1] = sortDirections_[2] = sortDirections_[3] = 0;

	pCompressFileListCtrl_->ClearAll();
	pCompressFileListCtrl_->InsertColumn(0, wxT("File name"));
	pCompressFileListCtrl_->InsertColumn(1, wxT("File size"));
	pCompressFileListCtrl_->InsertColumn(2, wxT("Compressed size"));

	pUncompressFileListCtrl_->ClearAll();
	pUncompressFileListCtrl_->InsertColumn(0, wxT("File name"));
	pUncompressFileListCtrl_->InsertColumn(1, wxT("File size"));
	pUncompressFileListCtrl_->InsertColumn(2, wxT("Uncompressed size"));

	pSaveCompressedFilesChoice_->Clear();
	pSaveCompressedFilesChoice_->Append(wxT("Save compressed file next to original (suffix .pjg)"));
	pSaveCompressedFilesChoice_->Append(wxT("Save compressed file to new directory"));
	pSaveCompressedFilesChoice_->SetSelection(0);
	pOutputDirPicker_->Disable();
#if defined(__WXOSX__)
	// Workaround for different behaviour of OS X port
	pOutputDirPicker_->GetTextCtrl()->Disable();
	pOutputDirPicker_->GetPickerCtrl()->Disable();
#endif
	pCompressedFilesBoxSizer_->Layout();

	pSaveUncompressedFilesChoice_->Clear();
	pSaveUncompressedFilesChoice_->Append(wxT("Save uncompressed file next to compressed"));
	pSaveUncompressedFilesChoice_->Append(wxT("Save uncompressed file to new directory"));
	pSaveUncompressedFilesChoice_->SetSelection(0);
	pUOutputDirPicker_->Disable();
#if defined(__WXOSX__)
	// Workaround for different behaviour of OS X port
	pUOutputDirPicker_->GetTextCtrl()->Disable();
	pUOutputDirPicker_->GetPickerCtrl()->Disable();
#endif
	pUncompressedFilesBoxSizer_->Layout();

	pProgressGauge_->Disable();
	pAbortCompressionButton_->Disable();
	pAbortDecompressionButton_->Disable();

//	InitDialog();	// Initialize validators

	SetIcon(getIcon());

	pDropTarget_ = new PJDropTarget();
	pDropTarget_->setMainFrame(this);
	SetDropTarget(pDropTarget_);	// Memory responsibility goes to wxWindow

#if defined(PJ_WIN32)
	if(calledFromCOM_OLE)
	{
		pPJDropTargetWin32_ = new PJDropTargetWin32();
		pPJDropTargetWin32_->setMainFrame(this);
	}
#endif
	pPJWorkerThreadManager_ = new PJWorkerThreadManager();
	previewGeneratorThread_.startPreviewThread();

	// Workaround for crash on *nix/GTK: Connect event here, not in the event table
	pMainNotebook_->Connect( wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGING, wxNotebookEventHandler( wxPackJPGMainFrame::OnMainNotebookPageChanging ), NULL, this );

	// Workaround for OS X: Hide empty menus
#if defined(__WXMAC__) || defined(__WXOSX__) || defined(__WXOSX_COCOA__)
	CocoaBridge::hideEmptyMenus();
#endif

#if defined(PJ_WIN32)
	pMainNotebook_->SetBackgroundColour( wxSystemSettings::GetColour( wxSYS_COLOUR_FRAMEBK ) );
#endif

	aTimer_.Start(50);
}

wxPackJPGMainFrame::~wxPackJPGMainFrame()
{
	pMainNotebook_->Disconnect( wxEVT_COMMAND_NOTEBOOK_PAGE_CHANGING, wxNotebookEventHandler( wxPackJPGMainFrame::OnMainNotebookPageChanging ), NULL, this );

#if defined(PJ_WIN32)
	delete pPJDropTargetWin32_;
#endif
	delete pPJWorkerThreadManager_;
	aTimer_.Stop();
}

bool wxPackJPGMainFrame::addDragAndDropFiles(const wxArrayString &filenames)
{
	int currentPageId = pMainNotebook_->GetCurrentPage()->GetId();

	int rejectedFiles = 0;
	for(size_t i = 0; i < filenames.GetCount(); i++)
	{
		bool isFileAccepted = false;
		if(currentPageId == ID_COMPRESSIONPANEL)
			isFileAccepted = addFileToCompressList(filenames[i]);
		else if(currentPageId == ID_DECOMPRESSIONPANEL)
			isFileAccepted = addFileToUncompressList(filenames[i]);

		if(!isFileAccepted)
			rejectedFiles++;
	}

	if(rejectedFiles > 0)
	{
		// Don't show message box now, because at least on Windows
		// the dropped files remain visible during the message box
		// which looks like the application is "broken".
		// Solution: Post message that will be processed later in the GUI message loop
		if(currentPageId == ID_COMPRESSIONPANEL)
		{
			// We don't need the thread-safe QueueEvent() here as we are still in the GUI thread
			wxCommandEvent showMsgBoxEvent(myCustomEventType, ID_SHOW_FILETYPE_WARNING_JPEG);
			this->AddPendingEvent(showMsgBoxEvent);
		}
		else if(currentPageId == ID_DECOMPRESSIONPANEL)
		{
			wxCommandEvent showMsgBoxEvent(myCustomEventType, ID_SHOW_FILETYPE_WARNING_PJG);
			this->AddPendingEvent(showMsgBoxEvent);
		}
	}

	return true;
}

bool wxPackJPGMainFrame::addShellExtFiles(const wxArrayString &filenames)
{
	// Check extension of the first file, then activate the respective page in the notebook
	if(filenames.size() == 0)
		return true;
	int currentPageId = pMainNotebook_->GetCurrentPage()->GetId();
	wxFileName fn(filenames[0]);

	wxString ext = fn.GetExt();
	if(fn.HasExt() && (fn.GetExt().CmpNoCase(_T("PJG")) == 0))
	{
		// Packed JPEG
		if(currentPageId == ID_COMPRESSIONPANEL)
		{
			pMainNotebook_->SetSelection(1);
		}
	}
	else
	{
		// JPEG
		if(currentPageId == ID_DECOMPRESSIONPANEL)
		{
			pMainNotebook_->SetSelection(0);
		}
	}

	return addDragAndDropFiles(filenames);
}

bool wxPackJPGMainFrame::addFileToCompressList(const wxString &filename)
{
	// Check file type
	PJFileType::FileType fileType = PJFileType::determineFileType(filename);
	if(fileType != PJFileType::FileTypeJPEG)
		return false;

	// Check whether file is already in the list
	long newId = pCompressFileListCtrl_->GetItemCount();
	wxFileName fn(filename);
	fn.Normalize();
	wxString fnString1 = fn.GetFullPath();

	for(long i = 0; i < newId; i++)
	{
#if defined(PJ_WIN32)
		// The code below is too slow on Windows
		wxString fnString2 = pCompressFileListCtrl_->GetItemText(i);
		if(fnString1.IsSameAs(fnString2, false))
			return true;
#else
		wxFileName fn2(pCompressFileListCtrl_->GetItemText(i));
		if(fn2.SameAs(fn))
			return true;	// Only return false to tell user about wrong file type
#endif
	}

	if(fn.FileExists())
	{
		pCompressFileListCtrl_->InsertItem(newId, filename);
		pCompressFileListCtrl_->SetItem(newId, 1, fn.GetHumanReadableSize());

		int listID = nextListID_[0]++;						// Get unique ID
		pCompressFileListCtrl_->SetItemData(newId, listID);	// Is used for sorting

		if(static_cast<int>(sortFileSizesC_.size()) < listID + 1)
			sortFileSizesC_.resize(listID + 1);
		sortFileSizesC_[listID] = static_cast<size_t>(fn.GetSize().GetValue());

		// Get some image informations
/*		{	// Currently commented out because it is slow and does not help the user
			// Better to give the user a preview of the selected image
			static const int THUMBNAIL_WIDTH = 320;
			static const int THUMBNAIL_HEIGHT = 240;

			wxImage image;
			image.SetOption(wxIMAGE_OPTION_MAX_WIDTH, THUMBNAIL_WIDTH);
			image.SetOption(wxIMAGE_OPTION_MAX_HEIGHT, THUMBNAIL_HEIGHT);

			wxString imageSizeStr(wxT("Not a JPEG file"));
			if ( image.LoadFile(filename) )
			{
				if(image.GetType() == wxBITMAP_TYPE_JPEG)
				{
					int origWidth = image.GetWidth();
					int origHeight = image.GetHeight();
					if(image.HasOption( wxIMAGE_OPTION_ORIGINAL_WIDTH )
						&& image.HasOption( wxIMAGE_OPTION_ORIGINAL_HEIGHT ))
					{
						origWidth = image.GetOptionInt( wxIMAGE_OPTION_ORIGINAL_WIDTH );
						origHeight = image.GetOptionInt( wxIMAGE_OPTION_ORIGINAL_HEIGHT );
					}
					imageSizeStr.Printf(wxT("%d x %d"), origWidth, origHeight);
				}
			}

			pCompressFileListCtrl_->SetItem(newId, 2, imageSizeStr);

		}*/
	}

	pCompressFileListCtrl_->SetColumnWidth(0, wxLIST_AUTOSIZE);
	pCompressFileListCtrl_->SetColumnWidth(1, wxLIST_AUTOSIZE);
	pCompressFileListCtrl_->SetColumnWidth(2, wxLIST_AUTOSIZE);

	return true;
}

bool wxPackJPGMainFrame::addFileToUncompressList(const wxString &filename)
{
	// Check file type
	PJFileType::FileType fileType = PJFileType::determineFileType(filename);
	if(fileType != PJFileType::FileTypePJG)
		return false;

	// Check whether file is already in the list
	long newId = pUncompressFileListCtrl_->GetItemCount();
	wxFileName fn(filename);
	fn.Normalize();
	wxString fnString1 = fn.GetFullPath();

	for(long i = 0; i < newId; i++)
	{
#if defined(PJ_WIN32)
		// wxFileName::SameAs is too slow on Windows, so use string comparison
		wxString fnString2 = pUncompressFileListCtrl_->GetItemText(i);
		if(fnString1.IsSameAs(fnString2, false))
			return true;
#else
		wxFileName fn2(pUncompressFileListCtrl_->GetItemText(i));
		if(fn2.SameAs(fn))
			return true;	// Only return false to tell user about wrong file type
#endif
	}

	if(fn.FileExists())
	{
		pUncompressFileListCtrl_->InsertItem(newId, filename);
		pUncompressFileListCtrl_->SetItem(newId, 1, fn.GetHumanReadableSize());

		int listID = nextListID_[1]++;							// Get unique ID
		pUncompressFileListCtrl_->SetItemData(newId, listID);	// Is used for sorting

		if(static_cast<int>(sortFileSizesUC_.size()) < newId + 1)
			sortFileSizesUC_.resize(newId + 1);
		sortFileSizesUC_[listID] = static_cast<size_t>(fn.GetSize().GetValue());
	}

	pUncompressFileListCtrl_->SetColumnWidth(0, wxLIST_AUTOSIZE);
	pUncompressFileListCtrl_->SetColumnWidth(1, wxLIST_AUTOSIZE);

	return true;
}

void wxPackJPGMainFrame::removeSelectedItems( wxListCtrl *pListCtrl )
{
	std::vector<int> itemsToBeDeleted;
	int stateMask = wxLIST_STATE_SELECTED;
	int itemCount = pListCtrl->GetItemCount();
	for(int i = 0; i < itemCount; i++)
	{
		int itemState = pListCtrl->GetItemState(i, stateMask);
		if(itemState != 0)
		{
			// Store items to be deleted in an array
			itemsToBeDeleted.push_back(i);
		}
	}

	// First, convert size_t to signed int to allow - 1 if list is empty
	int itemsToBeDeletedSize = static_cast<int>(itemsToBeDeleted.size());
	// Iterate list, start at the back
	for(int i = 0; i < itemsToBeDeletedSize; i++)
	{
		int itemToBeDeleted = itemsToBeDeleted[itemsToBeDeletedSize - 1 - i];
		pListCtrl->DeleteItem(itemToBeDeleted);
	}
}

void wxPackJPGMainFrame::addItemsToThreadManager(wxListCtrl *pListCtrl, bool outputToPath,
	bool overwriteFiles, bool verify,
	const wxString &outPath, const wxString &newFileExt)
{
	wxFileName outputDir(outPath, wxT(""));	// Only directory without file name

	long itemCount = pListCtrl->GetItemCount();
	long addedItems = 0;

	for(long i = 0; i < itemCount; i++)
		pListCtrl->SetItem(i, 2, wxEmptyString);

	pPJWorkerThreadManager_->clearWorkItems();

	for(long i = 0; i < itemCount; i++)
	{
		bool skipThisFile = false;
		wxString errorMessage;
		wxFileName fn(pListCtrl->GetItemText(i));

		wxFileName outFilename(fn);
		if(outputToPath)
		{
			outFilename = outputDir;
			outFilename.SetName(fn.GetName());
			outFilename.SetExt(newFileExt);
		}
		else
		{
			outFilename.SetExt(newFileExt);
		}

		outFilename.Normalize();

		// First, check whether outfile is already in the work items list
		// this can happen if two files with same names, but different endings are in the list of infiles
		bool isInOutList = pPJWorkerThreadManager_->existsOutfilename(outFilename.GetFullPath());

		// Check whether input = output (user added jpeg to decompression/pjg to compression)
		bool isSameFileInputOutput = outFilename.SameAs(fn);

		bool dirExists = outFilename.DirExists();
		bool canWriteFile = outFilename.IsDirWritable();

		if(!dirExists)
		{
			errorMessage = wxT("Error: Output directory does not exist");
			skipThisFile = true;
		}
		else if(!canWriteFile)
		{
			errorMessage = wxT("Error: Output directory is write protected");
			skipThisFile = true;
		}
		else if(isSameFileInputOutput)
		{
			errorMessage = wxT("Error: Output file is the same as input file");
			skipThisFile = true;
		}
		else if(isInOutList)
		{
			errorMessage = wxT("Error: Output file name is already used by another output file");
			skipThisFile = true;
		}

		// Check whether outFilename already exists
		if(!skipThisFile && !overwriteFiles)
		{
			// Then, check whether outfile exists
			bool outFileExists = outFilename.FileExists();

			if(outFileExists)
			{
				// Ask user
				PJOverwriteDialog dlg(this, outFilename.GetFullPath());
				int retVal = dlg.ShowModal();

				if(retVal == wxID_CANCEL)
				{
					// User clicked Cancel button: Clear all work items, return
					pPJWorkerThreadManager_->clearWorkItems();
					return;
				}
				else if(retVal == wxID_YESTOALL)
				{
					overwriteFiles = true;
				}
				else if(retVal == wxID_NO)
				{
					skipThisFile = true;
				}
			}
		}

		if(!skipThisFile)
		{
			pPJWorkerThreadManager_->addWorkItem(fn.GetFullPath(), outFilename.GetFullPath());
			addedItems++;
		}
		else
		{
			pListCtrl->SetItem(i, 2, errorMessage);
			pListCtrl->EnsureVisible(i);
			pListCtrl->SetColumnWidth(2, wxLIST_AUTOSIZE);
		}
	}

	if(addedItems > 0)
	{
		failedItemsCount_ = warningItemsCount_ = 0;
		bool ignoreWarnings = pIgnoreWarningsMenuItem_->IsChecked();
		bool discardMetadata = pDiscardMetadataMenuItem_->IsChecked();
		pPJWorkerThreadManager_->setParameters(verify, ignoreWarnings, discardMetadata, false);
		pPJWorkerThreadManager_->startWorking();
		enableUI(false);
		pUProgressGauge_->Enable();
		pUProgressGauge_->SetValue(0);
	}
}

void wxPackJPGMainFrame::enableUI(bool enable)
{
	pAboutMenuItem_->Enable(enable);
	pIgnoreWarningsMenuItem_->Enable(enable);
	pDiscardMetadataMenuItem_->Enable(enable);

	// If we do that, the Abort button is enabled but does not react on clicks...
//	enableSizer(pMainBoxSizer_, enable, true);

	// Workaround: Do not disable the notebook, but veto the Changing events
	enableSizer(pMainCompressionBoxSizer_, enable, true);
	enableSizer(pMainDecompressionBoxSizer_, enable, true);

	if(enable)
	{
		// Enable the OutputDirPicker depending on the SaveConvertedFilesChoice
		wxCommandEvent evt;
		OnSaveCompressedFilesChoice( evt );
		OnSaveUncompressedFilesChoice( evt );
	}

	pAbortCompressionButton_->Enable(!enable);	// Invert enable
	pAbortDecompressionButton_->Enable(!enable);	// Invert enable
}

void wxPackJPGMainFrame::startCommand( wxPackJPG_OLE_Commands cmd, const wxString &pathStr )
{
	// Instead of calling the methods directly, create an event that will do it for us.
	// Reason: Decouple COM calls from our methods. Otherwise the Windows Explorer
	// will be frozen if a method is long running, e.g. it displays a dialog.

	closeAfterCommandFinished_ = true;	// Make sure wxPackJPG closes after completing the command

	if(cmd == wxPJCmd_Compress_Here)
	{
#if wxCHECK_VERSION(2, 9, 0)
		wxCommandEvent* newEvent = new wxCommandEvent(wxEVT_BUTTON, ID_COMPRESSBUTTON);
		wxQueueEvent(this, newEvent);
#else
		wxCommandEvent newEvent(wxEVT_COMMAND_BUTTON_CLICKED, ID_COMPRESSBUTTON);
		AddPendingEvent(newEvent);
#endif
	}
	else if(cmd == wxPJCmd_Decompress_Here)
	{
#if wxCHECK_VERSION(2, 9, 0)
		wxCommandEvent* newEvent = new wxCommandEvent(wxEVT_BUTTON, ID_DECOMPRESSBUTTON);
		wxQueueEvent(this, newEvent);
#else
		wxCommandEvent newEvent(wxEVT_COMMAND_BUTTON_CLICKED, ID_DECOMPRESSBUTTON);
		AddPendingEvent(newEvent);
#endif
	}
	else if(cmd == wxPJCmd_Compress_To_Folder)
	{
		wxFileName path = wxFileName::DirName(pathStr);
		path.AppendDir(wxT("compressed"));
		if(!path.DirExists())
		{
			// Create compressed directory if it does not yet exist
			bool isOK = path.Mkdir();
			if(!isOK)
			{
				wxMessageBox(wxT("Could not create compressed folder"),
					wxT("wxPackJPG error"), wxOK | wxICON_ERROR);
				return;
			}
		}

		// Set output path
		pOutputDirPicker_->SetPath(path.GetPath());
		pSaveCompressedFilesChoice_->SetSelection(1);

		// Update UI widgets
		wxCommandEvent tmpEvent;
		OnSaveCompressedFilesChoice( tmpEvent );

#if wxCHECK_VERSION(2, 9, 0)
		wxCommandEvent* newEvent = new wxCommandEvent(wxEVT_BUTTON, ID_COMPRESSBUTTON);
		wxQueueEvent(this, newEvent);
#else
		wxCommandEvent newEvent(wxEVT_COMMAND_BUTTON_CLICKED, ID_COMPRESSBUTTON);
		AddPendingEvent(newEvent);
#endif
	}
	else if(cmd == wxPJCmd_Decompress_To_Folder)
	{
		wxFileName path = wxFileName::DirName(pathStr);
		path.AppendDir(wxT("uncompressed"));
		if(!path.DirExists())
		{
			// Create compressed directory if it does not yet exist
			bool isOK = path.Mkdir();
			if(!isOK)
			{
				wxMessageBox(wxT("Could not create uncompressed folder"),
					wxT("wxPackJPG error"), wxOK | wxICON_ERROR);
				return;
			}
		}

		// Set output path
		pUOutputDirPicker_->SetPath(path.GetPath());
		pSaveUncompressedFilesChoice_->SetSelection(1);

		// Update UI widgets
		wxCommandEvent tmpEvent;
		OnSaveCompressedFilesChoice( tmpEvent );

#if wxCHECK_VERSION(2, 9, 0)
		wxCommandEvent* newEvent = new wxCommandEvent(wxEVT_BUTTON, ID_DECOMPRESSBUTTON);
		wxQueueEvent(this, newEvent);
#else
		wxCommandEvent newEvent(wxEVT_COMMAND_BUTTON_CLICKED, ID_DECOMPRESSBUTTON);
		AddPendingEvent(newEvent);
#endif
	}
	else if(cmd == wxPJCmd_Compress_Folder_To_PJG)
	{
#if wxCHECK_VERSION(2, 9, 0)
		wxCommandEvent* newEvent = new wxCommandEvent(myCustomEventType, ID_CMD_COMPRESS_FOLDER_TO_PJG);
		newEvent->SetString(pathStr);
		wxQueueEvent(this, newEvent);
#else
		wxCommandEvent newEvent(myCustomEventType, ID_CMD_COMPRESS_FOLDER_TO_PJG);
		newEvent.SetString(pathStr);
		AddPendingEvent(newEvent);
#endif
	}
	else if(cmd == wxPJCmd_Decompress_Folder_To_JPG)
	{
#if wxCHECK_VERSION(2, 9, 0)
		wxCommandEvent* newEvent = new wxCommandEvent(myCustomEventType, ID_CMD_DECOMPRESS_FOLDER_TO_JPG);
		newEvent->SetString(pathStr);
		wxQueueEvent(this, newEvent);
#else
		wxCommandEvent newEvent(myCustomEventType, ID_CMD_DECOMPRESS_FOLDER_TO_JPG);
		newEvent.SetString(pathStr);
		AddPendingEvent(newEvent);
#endif
	}
}

void wxPackJPGMainFrame::OnMainFrameClose( wxCloseEvent& WXUNUSED(event) )
{
	previewGeneratorThread_.stopPreviewThread();
	aTimer_.Stop();
	Destroy();
}

void wxPackJPGMainFrame::OnExitMenuItem( wxCommandEvent& WXUNUSED(event) )
{
	Close();
}

void wxPackJPGMainFrame::OnAboutMenuItem( wxCommandEvent& WXUNUSED(event) )
{
	wxAboutDialogInfo info;
	info.SetName(wxT(WX_PACK_JPG_NAME) wxT(" ") wxT(WX_PACK_JPG_ARCHITECTURE_STRING));
	info.SetVersion(wxString(WX_PACK_JPG_VERSION_STRING, *wxConvCurrent));
	info.SetIcon(getIcon());

	wxString descrString;
	descrString = wxT("This program compresses and decompresses JPEG images.\n");
	descrString.Append(wxT("Built with: ") wxT(WX_PACK_JPG_COMPILER) wxT(" ") wxT(WX_PACK_JPG_COMPILER_VERSION) wxT("\n"));
	descrString.Append(wxT("Linked with:\n - ") wxVERSION_STRING wxT("\n - "));
	descrString.Append(wxString(pjglib_short_name(), *wxConvCurrent));
	info.SetDescription(descrString);
	info.SetCopyright(wxT("(C) ") wxT(WX_PACK_JPG_COPYRIGHT_YEAR) wxT(" ") wxT(WX_PACK_JPG_COPYRIGHT_NAME));

	wxAboutBox(info);
}

void wxPackJPGMainFrame::OnMainNotebookPageChanging( wxNotebookEvent& event )
{
//#if defined(__WXOSX__)
	// Unfortunately, on OS X it does not seem to be possible to disable the notebook control itself.
	// Also, disabling the notebook on Windows has side effects.
	// This is a workaround.
	if(pPJWorkerThreadManager_->getIsWorking())
		event.Veto();
//#endif
	previewGeneratorThread_.cancelPreview();
}

void wxPackJPGMainFrame::OnCompressButton( wxCommandEvent& WXUNUSED(event) )
{
	wxFileName outputDir = pOutputDirPicker_->GetPath();
	wxString outputDirName = outputDir.GetFullPath();

	bool outputToPath = (pSaveCompressedFilesChoice_->GetSelection() != 0);

	bool overwriteFiles = pOverwriteExistingFilesCheckBox_->GetValue();
	bool verify = pVerifyCheckBox_->GetValue();

	addItemsToThreadManager(pCompressFileListCtrl_, outputToPath,
		overwriteFiles, verify, outputDirName, wxT("pjg"));
}

void wxPackJPGMainFrame::OnAbortCompressionButton( wxCommandEvent& event )
{
	pPJWorkerThreadManager_->cancelWorking();
	pAbortCompressionButton_->Disable();

	pStatusBar_->SetStatusText(wxT("Please wait..."));
}

void wxPackJPGMainFrame::OnSaveCompressedFilesChoice( wxCommandEvent& WXUNUSED(event) )
{
	if(pSaveCompressedFilesChoice_->GetSelection() == 0)
	{
		pOutputDirPicker_->Disable();
#if defined(__WXOSX__)
		// Workaround for different behaviour of OS X port
		pOutputDirPicker_->GetTextCtrl()->Disable();
		pOutputDirPicker_->GetPickerCtrl()->Disable();
#endif
	}
	else
	{
		pOutputDirPicker_->Enable();
#if defined(__WXOSX__)
		// Workaround for different behaviour of OS X port
		pOutputDirPicker_->GetTextCtrl()->Enable();
		pOutputDirPicker_->GetPickerCtrl()->Enable();
#endif
	}
}

void wxPackJPGMainFrame::OnAddFilesButton( wxCommandEvent& WXUNUSED(event) )
{
	wxFileDialog chooseFileDialog(this, wxT("Add JPEG files"),
		wxEmptyString, wxEmptyString, wxT("JPEG files (*.jpg;*.jpeg;*.JPG;*.JPEG)|*.jpg;*.jpeg;*.JPG;*.JPEG|All files(*.*)|*.*"),
		wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_MULTIPLE);

	if(chooseFileDialog.ShowModal() == wxID_OK)
	{
		wxString directory = chooseFileDialog.GetDirectory();

		wxArrayString filenames;
		chooseFileDialog.GetFilenames(filenames);
		int rejectedFiles = 0;

		for(size_t i = 0; i < filenames.GetCount(); i++)
		{
			wxFileName fn(directory, filenames[i], wxPATH_NATIVE);
			wxString filename = fn.GetFullPath();
			PJFileType::FileType fileType = PJFileType::determineFileType(filename);
			if(fileType == PJFileType::FileTypeJPEG)
				addFileToCompressList(filename);
			else
				rejectedFiles++;
		}

		if(rejectedFiles > 0)
		{
			wxMessageBox(wxT("Some files were not added because they were not in the JPEG format."),
				wxT("Rejected files"), wxOK | wxICON_WARNING);
		}
	}
}

void wxPackJPGMainFrame::OnRemoveSelectedFilesButton( wxCommandEvent& WXUNUSED(event) )
{
	removeSelectedItems( pCompressFileListCtrl_ );
}

void wxPackJPGMainFrame::OnClearFileListButton( wxCommandEvent& WXUNUSED(event) )
{
	pCompressFileListCtrl_->DeleteAllItems();
	sortFileSizesC_.clear();
	nextListID_[0] = 0;
	sortDirections_[0] = sortDirections_[1] = 0;
	pPreviewCanvas_->clearPreviewImage();
}

// returns 0 if the items are equal, negative value if the first item is less than the second one and positive value if the first one is greater than the second one 
#if wxCHECK_VERSION(2, 9, 0)
int wxCALLBACK wxPackJPGMainFrame::wxListCompareFunction(wxIntPtr item1, wxIntPtr item2, wxIntPtr WXUNUSED(sortData))
#else
int wxCALLBACK wxPackJPGMainFrame::wxListCompareFunction(long item1, long item2, long WXUNUSED(sortData))
#endif
{
	wxListCtrl *pListCtrl = pSortListCtrl_;
	int col = sortColumn_;

	long index1 = pListCtrl->FindItem(-1, item1);	// Find list entry with data item1
	long index2 = pListCtrl->FindItem(-1, item2);
	bool reverseOrder = (sortDirections_[col + sortDirectionOffset_] == 1);

	if(col == 0)
	{
		wxString text1 = pListCtrl->GetItemText(index1);
		wxString text2 = pListCtrl->GetItemText(index2);

		int cmp = text1.Cmp(text2);
		if(reverseOrder)
			cmp = -cmp;
		return cmp;
	}
	else if(col == 1)
	{
		size_t fileSize1 = (*pSortFileSizes_)[item1];
		size_t fileSize2 = (*pSortFileSizes_)[item2];
		int cmp = 0;
		if(fileSize1 < fileSize2)
			cmp = -1;
		else if(fileSize1 > fileSize2)
			cmp = 1;
		if(reverseOrder)
			cmp = -cmp;
		return cmp;
	}

	return 0;
}

void wxPackJPGMainFrame::OnCompressFileListColClick( wxListEvent& event )
{
	int col = event.GetColumn();
	if(col < 2)
	{
		sortColumn_ = col;	// Store column in a static member for wxListCompareFunction
		pSortListCtrl_ = pCompressFileListCtrl_;	// Store list control
		sortDirectionOffset_ = 0;
		pSortFileSizes_ = &sortFileSizesC_;
		pCompressFileListCtrl_->SortItems( wxListCompareFunction, 0 );

		sortDirections_[col] = 1 - sortDirections_[col];
	}
}

void wxPackJPGMainFrame::OnCompressFileListItemDeselected( wxListEvent& event )
{
	pPreviewCanvas_->clearPreviewImage();
}

void wxPackJPGMainFrame::OnCompressFileListItemSelected( wxListEvent& event )
{
	long index = event.GetIndex();

	wxString filename = pCompressFileListCtrl_->GetItemText(index);
	previewGeneratorThread_.addPreviewRequest(filename);
}

void wxPackJPGMainFrame::OnCompressFileListKeyDown( wxListEvent& event )
{
#if defined(__WXOSX__)
	// Workaround for OS X: Ignore keys while working
	if(pPJWorkerThreadManager_->getIsWorking())
		return;
#endif
	int keyCode = event.GetKeyCode();
	if(keyCode == WXK_DELETE || keyCode == WXK_BACK)
	{
		// Delete selected items
		removeSelectedItems( pCompressFileListCtrl_ );
	}
}

void wxPackJPGMainFrame::OnDecompressButton( wxCommandEvent& WXUNUSED(event) )
{
	wxFileName outputDir = pUOutputDirPicker_->GetPath();
	wxString outputDirName = outputDir.GetFullPath();

	bool outputToPath = (pSaveUncompressedFilesChoice_->GetSelection() != 0);
	bool overwriteFiles = pUOverwriteExistingFilesCheckBox_->GetValue();

	addItemsToThreadManager(pUncompressFileListCtrl_, outputToPath,
		overwriteFiles, false, outputDirName, wxT("jpg"));
}

void wxPackJPGMainFrame::OnAbortDecompressionButton( wxCommandEvent& event )
{
	pPJWorkerThreadManager_->cancelWorking();
	pAbortDecompressionButton_->Disable();

	pStatusBar_->SetStatusText(wxT("Please wait..."));
}

void wxPackJPGMainFrame::OnSaveUncompressedFilesChoice( wxCommandEvent& WXUNUSED(event) )
{
	if(pSaveUncompressedFilesChoice_->GetSelection() == 0)
	{
		pUOutputDirPicker_->Disable();
#if defined(__WXOSX__)
		// Workaround for different behaviour of OS X port
		pUOutputDirPicker_->GetTextCtrl()->Disable();
		pUOutputDirPicker_->GetPickerCtrl()->Disable();
#endif
	}
	else
	{
		pUOutputDirPicker_->Enable();
#if defined(__WXOSX__)
		// Workaround for different behaviour of OS X port
		pUOutputDirPicker_->GetTextCtrl()->Enable();
		pUOutputDirPicker_->GetPickerCtrl()->Enable();
#endif
	}
}

void wxPackJPGMainFrame::OnUCAddFilesButton( wxCommandEvent& WXUNUSED(event) )
{
	wxFileDialog chooseFileDialog(this, wxT("Add Packed JPEG files"),
		wxEmptyString, wxEmptyString, wxT("Packed JPEG files (*.pjg)|*.pjg|All Files (*.*)|*.*"),
		wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_MULTIPLE);

	if(chooseFileDialog.ShowModal() == wxID_OK)
	{
		wxString directory = chooseFileDialog.GetDirectory();

		wxArrayString filenames;
		chooseFileDialog.GetFilenames(filenames);
		int rejectedFiles = 0;

		for(size_t i = 0; i < filenames.GetCount(); i++)
		{
			wxFileName fn(directory, filenames[i], wxPATH_NATIVE);
			wxString filename = fn.GetFullPath();
			PJFileType::FileType fileType = PJFileType::determineFileType(filename);
			if(fileType == PJFileType::FileTypePJG)
				addFileToUncompressList(filename);
			else
				rejectedFiles++;
		}

		if(rejectedFiles > 0)
		{
			wxMessageBox(wxT("Some files were not added because they were not in the PackJPG format."),
				wxT("Rejected files"), wxOK | wxICON_WARNING);
		}
	}
}

void wxPackJPGMainFrame::OnUCRemoveSelectedFilesButton( wxCommandEvent& WXUNUSED(event) )
{
	removeSelectedItems( pUncompressFileListCtrl_ );
}

void wxPackJPGMainFrame::OnUCClearFileListButton( wxCommandEvent& WXUNUSED(event) )
{
	pUncompressFileListCtrl_->DeleteAllItems();
	sortFileSizesUC_.clear();
	nextListID_[1] = 0;
	sortDirections_[2] = sortDirections_[3] = 0;
	pUCPreviewCanvas_->clearPreviewImage();
}

void wxPackJPGMainFrame::OnUCompressFileListColClick( wxListEvent& event )
{
	int col = event.GetColumn();
	if(col < 2)
	{
		sortColumn_ = col;	// Store column in a static member for wxListCompareFunction
		pSortListCtrl_ = pUncompressFileListCtrl_;	// Store list control
		sortDirectionOffset_ = 2;
		pSortFileSizes_ = &sortFileSizesUC_;
		pUncompressFileListCtrl_->SortItems( wxListCompareFunction, 0 );

		sortDirections_[col + 2] = 1 - sortDirections_[col + 2];
	}
}

void wxPackJPGMainFrame::OnUCompressFileListItemDeselected( wxListEvent& event )
{
	pUCPreviewCanvas_->clearPreviewImage();
}

void wxPackJPGMainFrame::OnUCompressFileListItemSelected( wxListEvent& event )
{
	long index = event.GetIndex();

	wxString filename = pUncompressFileListCtrl_->GetItemText(index);
	previewGeneratorThread_.addPreviewRequest(filename);
}

void wxPackJPGMainFrame::OnUCompressFileListKeyDown( wxListEvent& event )
{
#if defined(__WXOSX__)
	// Workaround for OS X: Ignore keys while working
	if(pPJWorkerThreadManager_->getIsWorking())
		return;
#endif
	int keyCode = event.GetKeyCode();
	if(keyCode == WXK_DELETE || keyCode == WXK_BACK)
	{
		// Delete selected items
		removeSelectedItems( pUncompressFileListCtrl_ );
	}
}

void wxPackJPGMainFrame::OnShowFileTypeWarningJPEG( wxCommandEvent & WXUNUSED(event) )
{
	wxMessageBox(wxT("Some files were not added because they were not in the JPEG format."),
		wxT("Rejected files"), wxOK | wxICON_WARNING);
}

void wxPackJPGMainFrame::OnShowFileTypeWarningPJG( wxCommandEvent & WXUNUSED(event) )
{
	wxMessageBox(wxT("Some files were not added because they were not in the PackJPG format."),
		wxT("Rejected files"), wxOK | wxICON_WARNING);
}

void wxPackJPGMainFrame::OnShowProcessingErrors( wxCommandEvent &event )
{
	if(failedItemsCount_ > 0 && warningItemsCount_ == 0)
	{
		wxString errorString = wxString::Format(wxT("There were %d processing errors, please see the detailed messages in the file list."),
			failedItemsCount_);
		wxMessageBox(errorString, wxT("Processing errors"), wxOK | wxICON_ERROR);
	}
	else if(failedItemsCount_ == 0 && warningItemsCount_ > 0)
	{
		wxString errorString = wxString::Format(wxT("There were %d processing warnings, please see the detailed messages in the file list."),
			warningItemsCount_);
		wxMessageBox(errorString, wxT("Processing warnings"), wxOK | wxICON_ERROR);
	}
	else if(failedItemsCount_ > 0 && warningItemsCount_ > 0)
	{
		wxString errorString = wxString::Format(wxT("There were %d processing errors and %d warnings, please see the detailed messages in the file list."), 
			failedItemsCount_, warningItemsCount_);
		wxMessageBox(errorString, wxT("Processing errors and warnings"), wxOK | wxICON_ERROR);
	}

	if(calledFromCOM_OLE_)
	{
//		Close();	// If called from COM/OLE, close window after operation is finished
	}
}



void wxPackJPGMainFrame::OnCmdCompressFolderToPJG( wxCommandEvent &event )
{
	// Check whether the command is already running
	if(pPJDirTreeThread_ != NULL)
		return;

	// Make sure correct tab is active
	int currentPageId = pMainNotebook_->GetCurrentPage()->GetId();
	if(currentPageId == ID_DECOMPRESSIONPANEL)
		pMainNotebook_->SetSelection(0);

	failedItemsCount_ = warningItemsCount_ = 0;
	enableUI(false);
	dirTreeFiles_.clear();

	pPJDirTreeThread_ = new PJDirTreeThread();
	pPJDirTreeThread_->startJob(PJDirTreeThread::PJDTTJobTypeCompress, event.GetString());
}

void wxPackJPGMainFrame::OnCmdDecompressFolderToJPG( wxCommandEvent &event )
{
	// Check whether the command is already running
	if(pPJDirTreeThread_ != NULL)
		return;

	// Make sure correct tab is active
	int currentPageId = pMainNotebook_->GetCurrentPage()->GetId();
	if(currentPageId == ID_COMPRESSIONPANEL)
		pMainNotebook_->SetSelection(1);

	failedItemsCount_ = warningItemsCount_ = 0;
	enableUI(false);
	dirTreeFiles_.clear();

	pPJDirTreeThread_ = new PJDirTreeThread();
	pPJDirTreeThread_->startJob(PJDirTreeThread::PJDTTJobTypeDecompress, event.GetString());
}

/**
 * This method is called after the PJDirTreeThread is finished.
 *
 * Tasks:
 * - Check whether output files are present
 * - If necessary, ask user what to do
 * - Start worker threads
 */
void wxPackJPGMainFrame::OnStartToFolder( wxCommandEvent & WXUNUSED(event) )
{
	bool overwriteFiles = false;		// TODO: Get values from settings (registry?)
	bool verify = false, ignoreWarnings = false, discardMetadata = false, createDir = true;
	wxListCtrl *pFileListCtrl = pCompressFileListCtrl_;
	int currentPageId = pMainNotebook_->GetCurrentPage()->GetId();
	if(currentPageId == ID_DECOMPRESSIONPANEL)
		pFileListCtrl = pUncompressFileListCtrl_;

	long itemCount = pFileListCtrl->GetItemCount();
	long addedItems = 0;

	for(long i = 0; i < itemCount; i++)
		pFileListCtrl->SetItem(i, 2, wxEmptyString);

	pPJWorkerThreadManager_->clearWorkItems();

	for(long i = 0; i < itemCount; i++)
	{
		bool skipThisFile = false;
		wxString errorMessage, filenameStr(pFileListCtrl->GetItemText(i));
		wxFileName fn(filenameStr);

		// Find the output filename
		wxFileName outFilename;
		bool fnFound = false;
		std::list< PJDirTreeThread::FoundFile >::const_iterator iter = dirTreeFiles_.begin();
		while(iter != dirTreeFiles_.end()
			&& !fnFound)
		{
			PJDirTreeThread::FoundFile file = *iter;
			if(file.inFilename.Cmp(filenameStr) == 0)
			{
				fnFound = true;
				outFilename = wxFileName(file.outFilename);
			}

			iter++;
		}

		wxASSERT(fnFound);

		outFilename.Normalize();

		// First, check whether outfile is already in the work items list
		// this can happen if two files with same names, but different endings are in the list of infiles
		bool isInOutList = pPJWorkerThreadManager_->existsOutfilename(outFilename.GetFullPath());

		// Check whether input = output (user added jpeg to decompression/pjg to compression)
		bool isSameFileInputOutput = outFilename.SameAs(fn);

		if(isSameFileInputOutput)
		{
			errorMessage = wxT("Error: Output file is the same as input file");
			skipThisFile = true;
		}
		else if(isInOutList)
		{
			errorMessage = wxT("Error: Output file name is already used by another output file");
			skipThisFile = true;
		}

		// Check whether outFilename already exists
		if(!skipThisFile && !overwriteFiles)
		{
			// Then, check whether outfile exists
			bool outFileExists = outFilename.FileExists();

			if(outFileExists)
			{
				// Ask user
				PJOverwriteDialog dlg(this, outFilename.GetFullPath());
				int retVal = dlg.ShowModal();

				if(retVal == wxID_CANCEL)
				{
					// User clicked Cancel button: Close application
					pPJWorkerThreadManager_->clearWorkItems();
					Close();
					return;
				}
				else if(retVal == wxID_YESTOALL)
				{
					overwriteFiles = true;
				}
				else if(retVal == wxID_NO)
				{
					skipThisFile = true;
				}
			}
		}

		if(!skipThisFile)
		{
			pPJWorkerThreadManager_->addWorkItem(fn.GetFullPath(), outFilename.GetFullPath());
			addedItems++;
		}
		else
		{
			pFileListCtrl->SetItem(i, 2, errorMessage);
			pFileListCtrl->EnsureVisible(i);
			pFileListCtrl->SetColumnWidth(2, wxLIST_AUTOSIZE);
		}
	}

	dirTreeFiles_.clear();

	if(addedItems > 0)
	{
		failedItemsCount_ = warningItemsCount_ = 0;
		pPJWorkerThreadManager_->setParameters(verify, ignoreWarnings, discardMetadata, createDir);
		pPJWorkerThreadManager_->startWorking();
		pUProgressGauge_->Enable();
		pUProgressGauge_->SetValue(0);
	}
	else
	{
		// Nothing to do...
		Close();
	}

}

void wxPackJPGMainFrame::OnTimer( wxTimerEvent & WXUNUSED(event) )
{
	if(firstTimeOnTimer_)
	{
		firstTimeOnTimer_ = false;
		if(calledFromCOM_OLE_)
		{
			Raise();	// A program called from COM/OLE is created in the background
			// Unfortunately, Raise() by itself is not always sufficient, as an application
			// that is not in the foreground can't bring itself to the foreground.
			// Solution: The shell extension DLL asks our HWND and calls SetForegroundWindow.
		}
	}

	// Check for the dirtree thread
	if(pPJDirTreeThread_ != NULL)
	{
		// Get work items
		PJDirTreeThread::FoundFile file;
		while(pPJDirTreeThread_->getWorkFile(file))
		{
			dirTreeFiles_.push_back(file);

			wxListCtrl *pFileListCtrl = pCompressFileListCtrl_;
			int currentPageId = pMainNotebook_->GetCurrentPage()->GetId();
			if(currentPageId == ID_DECOMPRESSIONPANEL)
				pFileListCtrl = pUncompressFileListCtrl_;

			long newId = pFileListCtrl->GetItemCount();
			wxFileName fn(file.inFilename);
			pFileListCtrl->InsertItem(newId, file.inFilename);
			pFileListCtrl->SetItem(newId, 1, fn.GetHumanReadableSize());

			int listID = 0;
			
			if(currentPageId == ID_COMPRESSIONPANEL)
				nextListID_[0]++;					// Get unique ID
			else
				nextListID_[1]++;					// Get unique ID

			pFileListCtrl->SetItemData(newId, listID);		// Is used for sorting

			if(currentPageId == ID_COMPRESSIONPANEL)
			{
				if(static_cast<int>(sortFileSizesC_.size()) < listID + 1)
					sortFileSizesC_.resize(listID + 1);
				sortFileSizesC_[listID] = static_cast<size_t>(fn.GetSize().GetValue());
			}
			else
			{
				if(static_cast<int>(sortFileSizesUC_.size()) < listID + 1)
					sortFileSizesUC_.resize(listID + 1);
				sortFileSizesUC_[listID] = static_cast<size_t>(fn.GetSize().GetValue());
			}

			pFileListCtrl->SetColumnWidth(0, wxLIST_AUTOSIZE);
			pFileListCtrl->SetColumnWidth(1, wxLIST_AUTOSIZE);
			pFileListCtrl->SetColumnWidth(2, wxLIST_AUTOSIZE);
		}


		if(!pPJDirTreeThread_->IsRunning())
		{
			// Thread is finished, clean it up
			wxThread::ExitCode exitCode = pPJDirTreeThread_->Wait();
			delete pPJDirTreeThread_;
			pPJDirTreeThread_ = NULL;

			// Indirectly call OnStartToFolder
			wxCommandEvent newEvent(myCustomEventType, ID_START_TO_FOLDER);
			this->AddPendingEvent(newEvent);
			// Reason for separate method: It's a bad idea to use dialog boxes
			// in OnTimer(). They also have a message loop and OnTimer is called
			// again, sort of recursively...
		}
	}

	// Check for the worker threads
	if(pPJWorkerThreadManager_ != NULL)
	{
		if(pPJWorkerThreadManager_->getIsWorking())
		{
			wxGauge *pProgressGauge = pProgressGauge_;
			wxListCtrl *pFileListCtrl = pCompressFileListCtrl_;
			int currentPageId = pMainNotebook_->GetCurrentPage()->GetId();
			if(currentPageId == ID_DECOMPRESSIONPANEL)
			{
				pProgressGauge = pUProgressGauge_;
				pFileListCtrl = pUncompressFileListCtrl_;
			}

			bool isValidFinishedWorkItem, isOK = false;
			wxString inFilename, outFilename, processMsg;
			isValidFinishedWorkItem = pPJWorkerThreadManager_->readFinishedWorkItem(inFilename, outFilename, processMsg, isOK);
			while(isValidFinishedWorkItem)
			{
				// Find index of inFilename in pCompressFileListCtrl_
				int index = 0;
				long itemCount = pFileListCtrl->GetItemCount();
				wxFileName fn(inFilename);
				for(long i = 0; i < itemCount; i++)
				{
#if defined(PJ_WIN32)
					// The code below is too slow on Windows
					wxString fnString2 = pFileListCtrl->GetItemText(i);
					if(inFilename.IsSameAs(fnString2, false))
					{
						index = i;
					}
#else
					wxFileName fn2(pFileListCtrl->GetItemText(i));
					if(fn2.SameAs(fn))
					{
						index = i;
					}
#endif
				}

				wxString newFileSizeString(wxT("Error"));
				wxFileName outFN(outFilename);
				if(isOK && outFN.FileExists())
					newFileSizeString = outFN.GetHumanReadableSize();
				else if(!isOK)
					newFileSizeString.Append( wxString(wxT(" (")) + processMsg + wxString(wxT(")")) );

				pFileListCtrl->SetItem(index, 2, newFileSizeString);
				pFileListCtrl->EnsureVisible(index);
				pFileListCtrl->SetColumnWidth(2, wxLIST_AUTOSIZE);

				if(!isOK)
					failedItemsCount_++;

				isValidFinishedWorkItem = pPJWorkerThreadManager_->readFinishedWorkItem(inFilename, outFilename, processMsg, isOK);
			}

			double progress = pPJWorkerThreadManager_->getProgress();

			if(pPJWorkerThreadManager_->isFinishedWorking())
			{
				pPJWorkerThreadManager_->stopWorking();
				progress = 1.0;
				enableUI(true);
				pStatusBar_->SetStatusText(wxEmptyString);

				// If some items had errors or warnings, notify user
				if(failedItemsCount_ > 0 || warningItemsCount_ > 0)
				{
					closeAfterCommandFinished_ = false;	// Don't close application, we had errors
					wxCommandEvent showMsgBoxEvent(myCustomEventType, ID_SHOW_PROCESSING_ERRORS);
					this->AddPendingEvent(showMsgBoxEvent);
				}
				else if(closeAfterCommandFinished_)
				{
					Close();	// If a command was sent via COM/OLE, close window after operation is finished
				}
			}

			int progressVal = static_cast<int>(std::floor(progress * static_cast<double>(pProgressGauge->GetRange()) + 0.5));
			pProgressGauge->SetValue(progressVal);
		}
	}

	// Check for preview thread
	wxImage previewImage;
	wxString previewFilename;
	if(previewGeneratorThread_.getNewPreviewImage(previewImage, previewFilename))
	{
		int currentPageId = pMainNotebook_->GetCurrentPage()->GetId();
		if(currentPageId == ID_DECOMPRESSIONPANEL)
			pUCPreviewCanvas_->previewImage(previewImage);
		else
			pPreviewCanvas_->previewImage(previewImage);
	}
}

// Icon: From resource on Win32, from PNG otherwise
#if !defined(PJ_WIN32)
#include "res/png/app_icon_png.h"
#endif

wxIcon wxPackJPGMainFrame::getIcon()
{
#if defined(PJ_WIN32)
	wxIcon icon(wxT("aaaaaaaa"));
#else
	wxMemoryInputStream istream(wx_pack_jpg_png, sizeof(wx_pack_jpg_png));
	wxImage myimage_img(istream, wxBITMAP_TYPE_PNG);
	wxBitmap myimage_bmp(myimage_img);
	wxIcon icon;
	icon.CopyFromBitmap(myimage_bmp);
#endif

	return icon;
}

void wxPackJPGMainFrame::enableSizer(wxSizer *pSizer, bool enable, bool recursive)
{
	wxStaticBoxSizer *pSBSizer = dynamic_cast<wxStaticBoxSizer*>(pSizer);
	if(pSBSizer != NULL)
	{
		wxStaticBox *pStaticBox = pSBSizer->GetStaticBox();
		if(pStaticBox != NULL)
			pStaticBox->Enable(enable);
	}

	wxSizerItemList &itemList = pSizer->GetChildren();
	wxSizerItemList::iterator iter = itemList.begin();
	while(iter != itemList.end())
	{
		wxSizerItem *pSizerItem = *iter;

		wxSizer *pInnerSizer = pSizerItem->GetSizer();
		if(pInnerSizer != NULL && recursive)
			enableSizer(pInnerSizer, enable, recursive);

		wxWindow *pInnerWindow = pSizerItem->GetWindow();
		if(pInnerWindow != NULL)
		{
			pInnerWindow->Enable(enable);

#if defined(__WXOSX__)
			// Workaround for different behaviour of OS X port
			wxSplitterWindow *pSplitterWindow = dynamic_cast<wxSplitterWindow*>(pInnerWindow);
			if(recursive && pSplitterWindow != NULL)
			{
				wxWindow *pInnerWin1 = pSplitterWindow->GetWindow1();
				if(pInnerWin1 != NULL)
				{
					pInnerWin1->Enable(enable);
					wxSizer *pInnerSizer1 = pInnerWin1->GetSizer();
					if(pInnerSizer1 != NULL && recursive)
						enableSizer(pInnerSizer1, enable, recursive);
				}
				wxWindow *pInnerWin2 = pSplitterWindow->GetWindow1();
				if(pInnerWin1 != NULL)
				{
					pInnerWin1->Enable(enable);
					wxSizer *pInnerSizer1 = pInnerWin1->GetSizer();
					if(pInnerSizer1 != NULL && recursive)
						enableSizer(pInnerSizer1, enable, recursive);
				}
			}

			wxDirPickerCtrl *pDirPickerCtrl = dynamic_cast<wxDirPickerCtrl*>(pInnerWindow);
			if(recursive && pDirPickerCtrl != NULL)
			{
				pDirPickerCtrl->GetTextCtrl()->Enable(enable);
				pDirPickerCtrl->GetPickerCtrl()->Enable(enable);
			}
#endif
		}
		iter++;
	}
}

BEGIN_EVENT_TABLE( wxPackJPGMainFrame, wxPackJPGMainFrameBase )
	EVT_COMMAND(ID_SHOW_FILETYPE_WARNING_JPEG, myCustomEventType, wxPackJPGMainFrame::OnShowFileTypeWarningJPEG)
	EVT_COMMAND(ID_SHOW_FILETYPE_WARNING_PJG, myCustomEventType, wxPackJPGMainFrame::OnShowFileTypeWarningPJG)
	EVT_COMMAND(ID_SHOW_PROCESSING_ERRORS, myCustomEventType, wxPackJPGMainFrame::OnShowProcessingErrors)
	EVT_COMMAND(ID_CMD_COMPRESS_FOLDER_TO_PJG, myCustomEventType, wxPackJPGMainFrame::OnCmdCompressFolderToPJG)
	EVT_COMMAND(ID_CMD_DECOMPRESS_FOLDER_TO_JPG, myCustomEventType, wxPackJPGMainFrame::OnCmdDecompressFolderToJPG)
	EVT_COMMAND(ID_START_TO_FOLDER, myCustomEventType, wxPackJPGMainFrame::OnStartToFolder)
	EVT_TIMER(ID_TIMER, wxPackJPGMainFrame::OnTimer)
END_EVENT_TABLE()
